#pragma once
#include "hello_imgui/runner_params.h"
#include "hello_imgui/hello_imgui_screenshot.h"
#include "hello_imgui/internal/backend_impls/backend_window_helper/backend_window_helper.h"
#include "hello_imgui/internal/backend_impls/backend_window_helper/window_geometry_helper.h"
#include "hello_imgui/internal/backend_impls/backend_window_helper/window_autosize_helper.h"

#include <memory>

namespace HelloImGui
{

class AbstractRunner
{
   public:
    /// Step 0: Construct a concrete Runner (for example RunnerSdlOpenGl3 or RunnerGlfwOpenGl3)
    AbstractRunner(RunnerParams &params_);
    virtual ~AbstractRunner() = default;

    RunnerParams & params;

    /// Step 4.a: Call Run()
    virtual void Run();

    /// Step 4.b: Or implement your own Run using Setup()/Render()/TearDown()
    void Setup();
    void CreateFramesAndRender(int idxFrame);
    void RenderGui(int idxFrame);
    void TearDown();

   public:
    // Events for mobile devices
    void OnPause();
    void OnResume();
    void OnDestroy();
    void OnLowMemory();

    // For jupyter notebook, which displays a screenshot post execution
    ImageBuffer ScreenshotRgb() { return Impl_ScreenshotRgb(); }

    //
    // Dpi related methods
    //
    // DpiWindowSizeFactor() is the factor by which window size should be multiplied to get a similar visible size on different OSes.
    // It returns ApplicationScreenPixelPerInch / 96  under windows and linux. Under macOS, it will return 1.
    float DpiWindowSizeFactor();
    // If we want a font to visually render like a font size of 14 we need to multiply its size by this factor
    float DpiFontLoadingFactor();


   protected:
    friend std::string GlslVersion();

    //
    // The methods Impl_* are astract
    // and shall be overriden in the concrete implementations by derivates
    //
    virtual void Impl_InitBackend() = 0;
    virtual void Impl_Select_Gl_Version() = 0;
    virtual std::string Impl_GlslVersion() = 0;

    virtual void Impl_CreateWindow() = 0;
    virtual void Impl_CreateGlContext() = 0;
    virtual void Impl_InitGlLoader() = 0;
    virtual void Impl_SetupPlatformRendererBindings() = 0;
    virtual void Impl_SetupImgGuiContext() {}

    virtual bool Impl_PollEvents() = 0;  // returns true if exit required
    virtual void Impl_NewFrame_3D() = 0;
    virtual void Impl_NewFrame_Backend() = 0;
    virtual void Impl_Frame_3D_ClearColor() = 0;
    virtual void Impl_RenderDrawData_To_3D() = 0;
    virtual void Impl_UpdateAndRenderAdditionalPlatformWindows() = 0;
    virtual void Impl_SwapBuffers() = 0;
    virtual void Impl_Cleanup() = 0;

    virtual ImageBuffer Impl_ScreenshotRgb() { return ImageBuffer{}; }

private:
    void PrepareAutoSize();
    void FinishWindowSetupOnSecondFrame();
    void ReloadFontIfFailed();
    void FinishAutoSize_IfRequired();
    void MakeWindowSizeRelativeTo96Ppi_IfRequired();
    bool ShallSizeWindowRelativeTo96Ppi();

protected:
    BackendApi::WindowPointer mWindow = nullptr;
    std::unique_ptr<BackendApi::IBackendWindowHelper> mBackendWindowHelper;
private:
    std::unique_ptr<WindowGeometryHelper> mGeometryHelper;
    std::unique_ptr<WindowAutoSizeHelper> mAutoSizeHelper;
    bool mPotentialFontLoadingError = false;
};

}  // namespace HelloImGui
